@@ -30,6 +30,9 @@ import requests
 import math
 from inorbit_edge.utils import encode_floating_point_list
 import certifi
+import subprocess
+import re
+from deprecated import deprecated
 
 INORBIT_CLOUD_SDK_ROBOT_CONFIG_URL = "https://control.inorbit.ai/cloud_sdk_robot_config"
 
@@ -41,6 +44,7 @@ MQTT_SUBTOPIC_CUSTOM_DATA = "custom"
 MQTT_SUBTOPIC_CUSTOM_COMMAND = "custom_command"
 MQTT_SUBTOPIC_STATE = "state"
 MQTT_SUBTOPIC_CAMERA_V2 = "ros/camera2"
+MQTT_SUBTOPIC_OUT_CMD = "out_cmd"
 
 MQTT_TOPIC_ECHO = "echo"
 MQTT_NAV_GOAL_GOAL = "ros/loc/nav_goal"
@@ -141,6 +145,8 @@ class RobotSession:
 
         self.command_callbacks = []
         self.camera_streamers = {}
+        self.camera_streaming_on = False
+        self.camera_streaming_mutex = threading.Lock()
         self.message_handlers[MQTT_INITIAL_POSE] = self._handle_initial_pose
         self.message_handlers[MQTT_CUSTOM_COMMAND] = self._handle_custom_command
         self.message_handlers[MQTT_NAV_GOAL_GOAL] = self._handle_nav_goal
@@ -287,6 +293,8 @@ class RobotSession:
             topic=self._get_robot_subtopic(subtopic=MQTT_NAV_GOAL_GOAL)
         )
         self.client.subscribe(topic=self._get_robot_subtopic(subtopic=MQTT_IN_CMD))
+        # ask server to resend modules, so our state is consistent with the server side
+        self._resend_modules()
 
     def _on_message(self, client, userdata, msg):
         """MQTT client message callback.
@@ -408,13 +416,17 @@ class RobotSession:
 
     def _start_cameras_streaming(self):
         """Start streaming on all registered cameras"""
-        for s in self.camera_streamers.values():
-            s.start()
+        with self.camera_streaming_mutex:
+            self.camera_streaming_on = True
+            for s in self.camera_streamers.values():
+                s.start()
 
     def _stop_cameras_streaming(self):
         """Start streaming on all registered cameras"""
-        for s in self.camera_streamers.values():
-            s.stop()
+        with self.camera_streaming_mutex:
+            self.camera_streaming_on = False
+            for s in self.camera_streamers.values():
+                s.stop()
 
     def publish_camera_frame(self, camera_id, image, width, height, ts):
         """Publishes a camera frame"""
@@ -458,6 +470,34 @@ class RobotSession:
         msg.return_code = result_code
         self.publish_protobuf(MQTT_SCRIPT_OUTPUT_TOPIC, msg)
 
+    def register_commands_path(self, path="./user_scripts", exec_name_regex=r".*"):
+        """Registers executable commands that handle InOrbit custom command actions.
+        Use `exec_name_regex` and `path` to customize which executables can be
+        accessed. Note that if an action script name matches `exec_name_regex`, then
+        the program will be executed prepending the provided `path`.
+        """
+
+        def handler(command_name, args, options):
+            if command_name != COMMAND_CUSTOM_COMMAND:
+                return
+            script_name = args[0]
+            script_args = args[1]
+            if re.match(exec_name_regex, script_name):
+                # TODO(mike) handle script return and output
+                try:
+                    subprocess.Popen(
+                        [f"{path}/{script_name}"] + list(script_args),
+                        shell=False,
+                        env=dict(os.environ, INORBIT_ROBOT_ID=self.robot_id),
+                    )
+                except Exception as ex:
+                    self.logger.error(
+                        f"Failed to run executable command: {script_name} {ex}"
+                    )
+                    options["result_function"]("1")
+
+        self.register_command_callback(handler)
+
     def register_command_callback(self, callback):
         """Register a function to be called when a command for the robot is received.
 
@@ -506,6 +546,17 @@ class RobotSession:
             )
 
         self.camera_streamers[camera_id] = CameraStreamer(camera, publish)
+        with self.camera_streaming_mutex:
+            if self.camera_streaming_on:
+                self.camera_streamers[camera_id].start()
+
+    def _resend_modules(self):
+        """Ask server to resend modules"""
+        self.publish(
+            self._get_robot_subtopic(subtopic=MQTT_SUBTOPIC_OUT_CMD),
+            "resend_modules",
+            qos=1,
+        )
 
     def _send_robot_status(self, robot_status):
         """Sends robot online/offline status message.
@@ -880,13 +931,45 @@ class RobotSessionFactory:
         constructor of instances.
         """
         self.robot_session_kw_args = robot_session_kw_args
+        self.command_callbacks = []
+        self.commands_paths_rules = []
 
     def build(self, robot_id, robot_name):
         """Builds a RobotSession object using the provided id and name.
         It also passes the robot_session_kw_args set when creating the factory to the
         RobotSession constructor.
         """
-        return RobotSession(robot_id, robot_name, **self.robot_session_kw_args)
+        session = RobotSession(robot_id, robot_name, **self.robot_session_kw_args)
+
+        def build_callback(callback):
+            def c(*args):
+                callback(robot_id, *args)
+
+            return c
+
+        for command_callback in self.command_callbacks:
+            session.register_command_callback(build_callback(command_callback))
+
+        for path, exec_name_regex in self.commands_paths_rules:
+            session.register_commands_path(path, exec_name_regex)
+        return session
+
+    def register_command_callback(self, callback):
+        """Register a command callback to be used on all robot sessions created
+        by this factory"""
+        if not callable(callback):
+            # Don't do anything if callback is not a valid function
+            return
+
+        self.command_callbacks.append(callback)
+
+    def register_commands_path(self, path="./user_scripts", exec_name_regex=r".*"):
+        """Registers executable commands that handle InOrbit custom command actions.
+        Use `exec_name_regex` and `path` to customize which executables can be
+        accessed. Note that if an action script name matches `exec_name_regex`, then
+        the program will be executed prepending the provided `path`.
+        """
+        self.commands_paths_rules.append((path, exec_name_regex))
 
 
 class RobotSessionPool:
@@ -902,13 +985,6 @@ class RobotSessionPool:
         self.robot_session_factory = robot_session_factory
         self.robot_sessions = {}
         self.getting_session_mutex = threading.Lock()
-        self.command_callbacks = []
-
-    def dispatch_command(self, args):
-        """Relays a command received from a robot session to all command
-        callbacks registered by users."""
-        for command_callback in self.command_callbacks:
-            command_callback(*args)
 
     def get_session(self, robot_id, robot_name=""):
         """Returns a connected RobotSession for the specified robot"""
@@ -924,13 +1000,6 @@ class RobotSessionPool:
                     robot_id, robot_name
                 )
                 self.robot_sessions[robot_id].connect()
-                if new_robot_session:
-
-                    def callback(*args):
-                        self.dispatch_command([robot_id, *args])
-
-                    self.robot_sessions[robot_id].register_command_callback(callback)
-
             return self.robot_sessions[robot_id]
         finally:
             self.getting_session_mutex.release()
@@ -953,9 +1022,10 @@ class RobotSessionPool:
         sess.disconnect()
         del self.robot_sessions[robot_id]
 
+    @deprecated(
+        version="1.7.2",
+        reason="use RobotSessionFactory.register_command_callback() instead",
+    )
     def register_command_callback(self, callback):
-        if not callable(callback):
-            # Don't do anything if callback is not a valid function
-            return
-
-        self.command_callbacks.append(callback)
+        """Registers a callback to be called when a command is received"""
+        self.robot_session_factory.register_command_callback(callback)
